unit CpuUsageP;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls;

const
  _fxmax=2048;

type
  TPDWord=^DWORD;
  TSystem_Basic_Information=packed record
    dwUnknown1: DWORD;
    uKeMaximumIncrement: ULONG;
    uPageSize: ULONG;
    uMmNumberOfPhysicalPages: ULONG;
    uMmLowestPhysicalPage: ULONG;
    uMmHighestPhysicalPage: ULONG;
    uAllocationGranularity: ULONG;
    pLowestUserAddress: Pointer;
    pMmHighestUserAddress: Pointer;
    uKeActiveProcessors: ULONG;
    bKeNumberProcessors: byte;
    bUnknown2: byte;
    wUnknown3: word;
  end;

  TSystem_Performance_Information = packed record
    liIdleTime: LARGE_INTEGER; //LARGE_INTEGER
    dwSpare: array [0..75] of DWORD;
  end;

  TSystem_Time_Information = packed record
    liKeBootTime: LARGE_INTEGER;
    liKeSystemTime: LARGE_INTEGER;
    liExpTimeZoneBias: LARGE_INTEGER;
    uCurrentTimeZoneId: ULONG;
    dwReserved: DWORD;
  end;

  tcupstyle=(cupnone,cuppolyline,cupline,cuprectangle,cuppyramid);

  TCpuUsageP=class(Tcustompanel)
  private
    fyline:integer;
    fspeed:integer;
    fxmax:integer;
    fpencolor:tcolor;
    fpenwidth:integer;
    fbackcolor:tcolor;
    fbackpic:tpicture;
    fusage:integer;
    FOnCUP_Usage:TNotifyEvent;
    fstyle:tcupstyle;
    fgitterstepx:integer;
    fgitterstepy:integer;
    fgittercolor:tcolor;

    fpa:array[0.._fxmax+1]of tpoint;
    fbmp,fbackbmp:tbitmap;
    ftm:ttimer;
    fhold:integer;
    fxgitter:integer;

    NtQuerySystemInformation: function(
      infoClass: DWORD; buffer: Pointer; bufSize:
      DWORD; returnSize: TPDWord
    ):DWORD;stdcall;
    liOldIdleTime:LARGE_INTEGER;
    liOldSystemTime:LARGE_INTEGER;

    procedure pic2backbmp;
    procedure backpicChanged(Sender: TObject);
    procedure dotimer(sender:tobject);

    procedure setyline(v:integer);
    procedure setspeed(v:integer);
    procedure setxmax(v:integer);
    procedure setpencolor(v:tcolor);
    procedure setpenwidth(v:integer);
    procedure setbackcolor(v:tcolor);
    procedure setpic(v:tpicture);
    procedure setstyle(v:tcupstyle);
    procedure setgitterstepx(v:integer);
    procedure setgitterstepy(v:integer);
    procedure setgittercolor(v:tcolor);

    procedure mkgraf;
  public
    constructor create(aowner:tcomponent);override;
    destructor destroy;override;
    procedure paint;override;
    procedure resize;override;

    procedure start;
    procedure stopp;
    function GetCPUUsage:double;
  published
    property CUP_BackColor:tcolor read fbackcolor write setbackcolor;
    property CUP_YLine:integer read fyline write setyline;
    property CUP_PenColor:tcolor read fpencolor write setpencolor;
    property CUP_PenWidth:integer read fpenwidth write setpenwidth;
    property CUP_Speed:integer read fspeed write setspeed;
    property CUP_XMax:integer read fxmax write setxmax;
    property CUP_Pic:tpicture read fbackpic write setpic;
    property CUP_Usage:integer read fusage;
    property CUP_Style:TCupStyle read fstyle write setstyle;

    property CUP_GitterStepX:integer read fgitterstepx write setgitterstepx;
    property CUP_GitterStepY:integer read fgitterstepy write setgitterstepy;
    property CUP_GitterColor:Tcolor read fgittercolor write setgittercolor;

    property OnCUP_Usage:TNotifyEvent read FOnCUP_Usage write FOnCUP_Usage;


    property Align;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnCanResize;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDrag;
  end;


procedure Register;

implementation

{$R cpuusagep.res}

procedure Register;
begin
  RegisterComponents('DAN',[TCPUUsageP]);
end;

constructor TCpuUsageP.create(aowner:tcomponent);
var
  r:integer;
begin
  inherited;

  fhold:=height;

  fusage:=0;
  fxmax:=100;
  fyline:=2;
  fbackcolor:=clblack;
  fpencolor:=clgreen;
  fpenwidth:=1;
  fspeed:=100;
  fstyle:=cuppolyline;

  fgitterstepx:=10;
  fgitterstepy:=5;
  fgittercolor:=clnavy;


  fbmp:=tbitmap.create;
  fbmp.PixelFormat:=pf24bit;
  fbmp.width:=width;
  fbmp.Height:=height;

  fbackbmp:=tbitmap.create;
  fbackbmp.PixelFormat:=pf24bit;
  fbackbmp.width:=width;
  fbackbmp.Height:=height;

  //backbmp.LoadFromFile('c:\tmp\t.bmp');

  fbackpic:=tpicture.Create;
  fbackpic.OnChange:=backpicChanged;


  ftm:=ttimer.Create(self);
  ftm.OnTimer:=dotimer;
  ftm.Interval:=fspeed;
  ftm.Enabled:=false;

  fbmp.Canvas.Brush.color:=fbackcolor;
  fbmp.canvas.pen.color:=fpencolor;

  for r:=0 to _fxmax do fpa[r]:=point(0,fbmp.height-cup_yline);
end;

destructor TCpuUsageP.destroy;
begin
  ftm.Enabled:=false;
  ftm.Free;
  fbmp.Free;
  fbackbmp.Free;
  fbackpic.free;
  inherited;
end;

procedure TCpuUsageP.backpicChanged(Sender: TObject);
begin
  pic2backbmp;
end;

procedure TCpuUsageP.pic2backbmp;
var
  b:tbitmap;
  x,y,l,t,w,h:integer;
begin
  if fbackpic.graphic=nil then begin
    mkgraf;
    exit;
  end;

  try
    b:=tbitmap.create;
    b.PixelFormat:=pf24bit;

    try
      b.assign(fbackpic.Graphic);

      fbackbmp.width:=screen.width;
      fbackbmp.height:=screen.height;

      w:=b.width;
      h:=b.Height;

      t:=0;
      for y:=0 to trunc(screen.width/h) do begin
        l:=0;
        for x:=0 to trunc(screen.width/w) do begin
          bitblt(
            fbackbmp.Canvas.Handle,l,t,w,h,
            b.Canvas.Handle,0,0,
            srccopy
          );
          l:=l+w;
        end;
        t:=t+h;
      end;
    finally
      b.Free;
    end;

  except
    fbackpic.graphic:=nil;
  end;
  mkgraf;
end;

procedure TCpuUsageP.setpic(v:tpicture);
begin
  try
    fbackpic.assign(v);
    pic2backbmp;
  except
    fbackpic.graphic:=nil;
  end;
end;

procedure TCpuUsageP.setgittercolor(v:tcolor);
begin
  fgittercolor:=v;
  mkgraf;
end;

procedure TCpuUsageP.setgitterstepx(v:integer);
begin
  fgitterstepx:=v;
  mkgraf;
end;

procedure TCpuUsageP.setgitterstepy(v:integer);
begin
  fgitterstepy:=v;
  mkgraf;
end;

procedure TCpuUsageP.setstyle(v:tcupstyle);
begin
  fstyle:=v;
  mkgraf;
end;


procedure TCpuUsageP.setspeed(v:integer);
begin
  if v>=0 then begin
    fspeed:=v;
    ftm.Interval:=v;
  end;
end;

procedure TCpuUsageP.setxmax(v:integer);
var
  xmaxold,r:integer;
  dx:double;
begin
  if(v>0)and(v<=_fxmax) then begin

    xmaxold:=fxmax;
    dx:=width/v;
    for r:=0 to v-1 do begin
      if r=v-1 then fpa[r].x:=width else fpa[r].x:=round(dx*r);
      if r>=xmaxold then fpa[r].y:=height-fyline;
    end;
    fxmax:=v;
    mkgraf;
  end;
end;

procedure TCpuUsageP.setyline(v:integer);
var
  r:integer;
  h,fyold:integer;
begin
  if(v>=0)and(v<=height)then begin
    fyold:=fyline;

    //berechne linie neu
    for r:=0 to fxmax do begin
      h:=height-fyold-fpa[r].y;
      h:=height-v-h;
      if h<0 then h:=0;
      if h>(height-v) then h:=height-v;
      fpa[r].y:=h;
    end;
    fyline:=v;
    mkgraf;
  end;
end;

procedure TCpuUsageP.setpencolor(v:tcolor);
begin
  fpencolor:=v;
  mkgraf;
end;

procedure TCpuUsageP.setpenwidth(v:integer);
begin
  if(v>0)and(v<=20) then begin
    fpenwidth:=v;
    mkgraf;
  end;
end;

procedure TCpuUsageP.setbackcolor(v:tcolor);
begin
  fbackcolor:=v;
  mkgraf;
end;

//****************************************************************************
//Autor: Thomas Stutz
//URL: http://www.swissdelphicenter.ch/de/showcode.php?id=969
//...die CPU Auslastung unter Windows NT/2000/XP ermitteln?
//****************************************************************************
function TCpuUsageP.GetCPUUsage:double;
const
  _SystemBasicInformation=0;
  _SystemPerformanceInformation=2;
  _SystemTimeInformation=3;
var
  SysBaseInfo:TSystem_Basic_Information;
  SysPerfInfo:TSystem_Performance_Information;
  SysTimeInfo:TSystem_Time_Information;
  status:Longint;
  dbSystemTime:Double;
  dbIdleTime:Double;

  function Li2Double(x:LARGE_INTEGER):Double;
  begin
    Result:=x.HighPart*4.294967296E9+x.LowPart;
  end;

begin
  result:=0;
  if @NtQuerySystemInformation=nil then begin
    NtQuerySystemInformation:=GetProcAddress(
      GetModuleHandle('ntdll.dll'),'NtQuerySystemInformation'
    );
  end;

  // get number of processors in the system
  status:=NtQuerySystemInformation(
    _SystemBasicInformation,@SysBaseInfo,SizeOf(SysBaseInfo),nil
  );
  if status<>0 then Exit;

  // get new system time
  status:=NtQuerySystemInformation(
    _SystemTimeInformation,@SysTimeInfo,SizeOf(SysTimeInfo),nil
  );
  if status<>0 then Exit;

  // get new CPU's idle time
  status:=NtQuerySystemInformation(
    _SystemPerformanceInformation,@SysPerfInfo,SizeOf(SysPerfInfo),nil
  );
  if status<>0 then Exit;

  // if it's a first call - skip it
  if(liOldIdleTime.QuadPart<>0) then begin
    // CurrentValue = NewValue - OldValue
    dbIdleTime:=Li2Double(SysPerfInfo.liIdleTime)-Li2Double(liOldIdleTime);
    dbSystemTime:=Li2Double(SysTimeInfo.liKeSystemTime)-Li2Double(liOldSystemTime);

    // CurrentCpuIdle = IdleTime / SystemTime
    dbIdleTime:=dbIdleTime/dbSystemTime;

    // CurrentCpuUsage% = 100 - (CurrentCpuIdle * 100) /NumberOfProcessors
    dbIdleTime:=100.0-dbIdleTime*100.0/SysBaseInfo.bKeNumberProcessors+0.5;

    // Show Percentage
    result:=dbIdleTime;
    if result>100 then result:=100
  end;
  // store new CPU's idle and system time
  liOldIdleTime:=SysPerfInfo.liIdleTime;
  liOldSystemTime:=SysTimeInfo.liKeSystemTime;
end;
//****************************************************************************

procedure TCpuUsageP.dotimer(sender:tobject);
var
  p:tpoint;
  r,hd,li:integer;
  dx:double;
begin
  if not visible then exit;

  try
    hd:=fbmp.height-fyline-5;if hd<=0 then hd:=0;
    dx:=fbmp.width/fxmax;

    //punkte verschieben
    for r:=1 to fxmax-1 do begin
      fpa[r-1].x:=round((r-1)*dx);
      fpa[r-1].y:=fpa[r].y;
    end;

    //neuen punkt: cpu auslastung
    li:=round(getcpuusage);
    hint:='CPU-Auslastung: '+inttostr(li)+'%';
    fusage:=li;
    //if Assigned(FOnCUP_Usage) then FOnCUP_Usage(Self);

    p.x:=width;
    hd:=round((hd*li)/100);
    p.y:=fbmp.height-hd-fyline;
    fpa[fxmax-1]:=p;

  except
  end;
  mkgraf;
end;

procedure tCpuUsageP.start;
var
  p:tpoint;
  r:integer;
  dx:double;
begin
  //setze graf auf fyline/nulllinie
  dx:=width/fxmax;
  for r:=0 to fxmax-1 do begin
    if r=fxmax-1 then p.x:=width else p.x:=round(dx*r);
    p.y:=height-fyline;
    fpa[r]:=p;
  end;
  fxgitter:=0;
  ftm.enabled:=true;
end;

procedure tCpuUsageP.stopp;
begin
  ftm.enabled:=false;
end;

procedure tCpuUsageP.paint;
begin
  try
    bitblt(
      canvas.handle,0,0,width,height,
      fbmp.canvas.Handle,0,0,
      srccopy
    );
  except
  end;
end;

procedure tCpuUsageP.resize;
var
  h,r:integer;
  p:tpoint;
  dx:double;
begin

  //punkte anpssen an neue hhe/breite
  dx:=width/fxmax;
  for r:=0 to fxmax-1 do begin
    if r=fxmax-1 then p.x:=width else p.x:=round(dx*r);
    //h/hold=hh/height
    h:=fhold-fyline-fpa[r].y;
    p.Y:=round((h*height)/fhold);
    p.y:=height-fyline-p.Y;
    fpa[r]:=p;
  end;
  mkgraf;

  //merke alte hhe
  fhold:=height;if fhold<2 then fhold:=2;
end;

procedure TCpuUsageP.mkgraf;
var
  y,r,t,l,w,h:integer;
  dy:double;
begin
  fbmp.width:=width;
  fbmp.height:=height;
  fbmp.canvas.brush.color:=fbackcolor;
  fbmp.Canvas.Pen.color:=fpencolor;
  fbmp.Canvas.Pen.width:=fpenwidth;

  if(fstyle=cupnone)or(fbackpic.graphic=nil) then begin
    fbmp.Canvas.FillRect(rect(0,0,fbmp.width,fbmp.height));
  end
  else begin
    try
      bitblt(
        fbmp.canvas.handle,0,0,width,height,
        fbackbmp.canvas.Handle,0,0,
        srccopy
      );
    except
    end;
  end;

  //gitter
  fbmp.Canvas.Pen.color:=fgittercolor;
  fbmp.Canvas.Pen.width:=1;

  //y-gitter
  if fgitterstepy>0 then begin
    dy:=height/fgitterstepy;
    for y:=1 to fgitterstepy-1 do begin
      fbmp.Canvas.moveto(0,round(dy*y));
      fbmp.Canvas.lineto(width,round(dy*y));
    end;
  end;

  //x-gitter
  if fgitterstepx>0 then begin
    for r:=0 to fxmax-1 do begin
      if (r mod fgitterstepx)<>fxgitter then continue;
      fbmp.Canvas.moveto(fpa[r].x,0);
      fbmp.Canvas.lineto(fpa[r].x,height);
    end;
  end;
  if ftm.enabled then begin
    dec(fxgitter);if fxgitter<0 then fxgitter:=fgitterstepx-1;
  end;

  fbmp.Canvas.Pen.color:=fpencolor;
  fbmp.Canvas.Pen.width:=fpenwidth;


  //graf
  if fstyle=cuppolyline then fbmp.canvas.Polyline(Slice(fpa,fxmax))
  else if fstyle=cupline then begin
    h:=height;
    for r:=0 to fxmax-1 do begin
      t:=fpa[r].y;
      l:=fpa[r].X;
      fbmp.canvas.moveto(l,t);
      fbmp.canvas.LineTo(l,h);
    end;
  end
  else if fstyle=cuprectangle then begin
    fbmp.Canvas.Pen.width:=1;
    w:=fpenwidth;
    h:=height-fyline;
    for r:=0 to fxmax-1 do begin
      t:=fpa[r].y;
      l:=fpa[r].X;
      fbmp.Canvas.Rectangle(l,t,l+w,h);
    end;
  end
  else if fstyle=cuppyramid then begin
    fbmp.Canvas.Pen.width:=1;
    w:=fpenwidth;
    h:=height;
    for r:=0 to fxmax-1 do begin
      t:=fpa[r].y;
      l:=fpa[r].X;
      fbmp.canvas.moveto(l,h);
      fbmp.canvas.lineto(l+(w div 2),t);
      fbmp.canvas.LineTo(l+w,h);
    end;
  end;

  paint;
end;

end.

